// ChatServer.cpp : Implementation of WinMain


// Note: Proxy/Stub Information
//      To build a separate proxy/stub DLL, 
//      run nmake -f ChatServerps.mk in the project directory.

#include "stdafx.h"
#include "resource.h"
#include <initguid.h>
#include "ChatServer.h"

#include "ChatServer_i.c"
#include "ChatServerObj.h"


const DWORD dwTimeOut = 5000; // time for EXE to be idle before shutting down
const DWORD dwPause = 1000; // time to wait for threads to finish up

// Passed to CreateThread to monitor the shutdown event
static DWORD WINAPI MonitorProc(void* pv)
{
    CExeModule* p = (CExeModule*)pv;
    p->MonitorShutdown();
    return 0;
}

LONG CExeModule::Unlock()
{
    LONG l = CComModule::Unlock();
    if (l == 0)
    {
        bActivity = true;
        SetEvent(hEventShutdown); // tell monitor that we transitioned to zero
    }
    return l;
}

//Monitors the shutdown event
void CExeModule::MonitorShutdown()
{
    while (1)
    {
        WaitForSingleObject(hEventShutdown, INFINITE);
        DWORD dwWait=0;
        do
        {
            bActivity = false;
            dwWait = WaitForSingleObject(hEventShutdown, dwTimeOut);
        } while (dwWait == WAIT_OBJECT_0);
        // timed out
        if (!bActivity && m_nLockCnt == 0) // if no activity let's really bail
        {
#if _WIN32_WINNT >= 0x0400 & defined(_ATL_FREE_THREADED)
            CoSuspendClassObjects();
            if (!bActivity && m_nLockCnt == 0)
#endif
                break;
        }
    }
    CloseHandle(hEventShutdown);
    PostThreadMessage(dwThreadID, WM_QUIT, 0, 0);
}

bool CExeModule::StartMonitor()
{
    hEventShutdown = CreateEvent(NULL, false, false, NULL);
    if (hEventShutdown == NULL)
        return false;
    DWORD dwThreadID;
    HANDLE h = CreateThread(NULL, 0, MonitorProc, this, 0, &dwThreadID);
    return (h != NULL);
}

CExeModule _Module;

BEGIN_OBJECT_MAP(ObjectMap)
OBJECT_ENTRY(CLSID_ChatServerObj, CChatServerObj)
END_OBJECT_MAP()


LPCTSTR FindOneOf(LPCTSTR p1, LPCTSTR p2)
{
    while (p1 != NULL && *p1 != NULL)
    {
        LPCTSTR p = p2;
        while (p != NULL && *p != NULL)
        {
            if (*p1 == *p)
                return CharNext(p1);
            p = CharNext(p);
        }
        p1 = CharNext(p1);
    }
    return NULL;
}


//-------------------------------------------------------------------
// Register interface marshalers internally.
//-------------------------------------------------------------------
#include <comdef.h>
DWORD RegisterInterfaceMarshaler(REFIID riid)
{
    DWORD dwCookie=0;
    try {
        IUnknown *pUnk=0;
        HRESULT hr = ::DllGetClassObject(riid, IID_IUnknown, (void**)&pUnk);
        if (FAILED(hr)) throw _com_error(hr);
        hr = ::CoRegisterClassObject(riid, pUnk, 
                CLSCTX_INPROC_SERVER,
                REGCLS_MULTIPLEUSE, &dwCookie);
        if (FAILED(hr)) throw _com_error(hr);
        hr = ::CoRegisterPSClsid(riid, riid);
        if (FAILED(hr)) throw _com_error(hr);
    } catch (_com_error & er) {
        ::MessageBox(0, er.ErrorMessage(), "ChatServer", MB_OK|MB_SERVICE_NOTIFICATION|MB_ICONERROR);
    }

    return dwCookie;
}

void UnregisterInterfaceMarshaler(DWORD dwCookie)
{
    HRESULT hr = ::CoRevokeClassObject(dwCookie);
}

inline void Chk(HRESULT hr) 
{
   if (FAILED(hr)) _com_issue_error(hr);
}


/////////////////////////////////////////////////////////////////////////////
//
extern "C" int WINAPI _tWinMain(HINSTANCE hInstance, 
    HINSTANCE /*hPrevInstance*/, LPTSTR lpCmdLine, int /*nShowCmd*/)
{
    lpCmdLine = GetCommandLine(); //this line necessary for _ATL_MIN_CRT

#if _WIN32_WINNT >= 0x0400 & defined(_ATL_FREE_THREADED)
    HRESULT hRes = CoInitializeEx(NULL, COINIT_MULTITHREADED);
#else
    HRESULT hRes = CoInitialize(NULL);
#endif
    _ASSERTE(SUCCEEDED(hRes));
    _Module.Init(ObjectMap, hInstance, &LIBID_CHATSERVERLib);
    _Module.dwThreadID = GetCurrentThreadId();
    TCHAR szTokens[] = _T("-/");

    int nRet = 0;
    BOOL bRun = TRUE;
    LPCTSTR lpszToken = FindOneOf(lpCmdLine, szTokens);
    while (lpszToken != NULL)
    {
        if (lstrcmpi(lpszToken, _T("UnregServer"))==0)
        {
            _Module.UpdateRegistryFromResource(IDR_ChatServer, FALSE);
            nRet = _Module.UnregisterServer(TRUE);
            bRun = FALSE;
            break;
        }
        if (lstrcmpi(lpszToken, _T("RegServer"))==0)
        {
            _Module.UpdateRegistryFromResource(IDR_ChatServer, TRUE);
            nRet = _Module.RegisterServer(TRUE);
            bRun = FALSE;
            break;
        }
        lpszToken = FindOneOf(lpszToken, szTokens);
    }

    if (bRun)
    {

      // OUR CODE----End
      // Parsing command-line
      TCHAR szDiscussion[MAX_PATH]; 
      TCHAR szBroker[MAX_PATH]; 
      TCHAR szTokens[] = _T("#");
      LPCTSTR lpszDiscuss = FindOneOf(lpCmdLine, szTokens);
      if (lpszDiscuss==NULL) {
         MessageBox(0, "Command Line:\n\nChatServer   #DiscussCOM   #ChatBroker", 
           "Usage", MB_OK|MB_SERVICE_NOTIFICATION|MB_ICONINFORMATION);
         return nRet;
      }

      LPCTSTR lpszBroker = FindOneOf(lpszDiscuss, szTokens);
      if (lpszBroker==NULL) {
         char szComputerName[MAX_COMPUTERNAME_LENGTH+1];
         DWORD dwSize = MAX_COMPUTERNAME_LENGTH+1;
         ::GetComputerName(szComputerName, &dwSize);
         strcpy(szBroker, szComputerName);
         strcpy(szDiscussion, lpszDiscuss);
      } else {
         strcpy(szBroker, lpszBroker);
         // Reason for subtracting 2:  # and space
         strncpy(szDiscussion, lpszDiscuss, lpszBroker-lpszDiscuss-2);
         szDiscussion[lpszBroker-lpszDiscuss-2] = '\0';
      }

      // Register interface Marshaler
      DWORD dwBroker = RegisterInterfaceMarshaler(IID_IRegisterChatServer);
      DWORD dwServer = RegisterInterfaceMarshaler(IID_ISendMessage);
      DWORD dwClient = RegisterInterfaceMarshaler(IID_IReceiveMessage);
      // OUR CODE----End



        // Generated code...Begin
        _Module.StartMonitor();
#if _WIN32_WINNT >= 0x0400 & defined(_ATL_FREE_THREADED)
        hRes = _Module.RegisterClassObjects(CLSCTX_LOCAL_SERVER, 
            REGCLS_MULTIPLEUSE | REGCLS_SUSPENDED);
        _ASSERTE(SUCCEEDED(hRes));
        hRes = CoResumeClassObjects();
#else
        hRes = _Module.RegisterClassObjects(CLSCTX_LOCAL_SERVER, 
            REGCLS_MULTIPLEUSE);
#endif
        _ASSERTE(SUCCEEDED(hRes));
        // Generated code...End



      // OUR CODE----Begin
      HRESULT hr=S_OK;
      try {
         CComPtr<ISendMessage> pSend;
         hr = CoCreateInstance(CLSID_ChatServerObj, 0, CLSCTX_SERVER, 
            IID_ISendMessage, (void**)&pSend);
         Chk(hr);

         MULTI_QI mqi[] = { {&IID_IRegisterChatServer, 0, S_OK} };
         COSERVERINFO csi = {0, CComBSTR(szBroker), 0, 0};
         // CLSID for the broker object, copied from ChatBroker_i.c
         const CLSID CLSID_ChatBrokerObj = 
         {0x3284B5FF,0x6893,0x11D2,{0x87,0xCE,0x00,0x60,0x08,0xCD,0xD9,0xAE}};
         hr = CoCreateInstanceEx(CLSID_ChatBrokerObj, 0, CLSCTX_SERVER, 
            &csi, sizeof(mqi)/sizeof(mqi[0]), mqi);
         Chk(hr); Chk(mqi[0].hr);

         // Register the Chat Server with the Chat Broker.
         // This server will support the bstrDiscussion topic.
         CComPtr<IRegisterChatServer> 
            pBroker((IRegisterChatServer*)mqi[0].pItf);
         CComBSTR bstrDiscussion(szDiscussion);
         hr = pBroker->Register(pSend, bstrDiscussion);
         if (FAILED(hr))
            _com_issue_errorex(hr, pBroker, IID_IRegisterChatServer);

         // Run the server until a user hits the OK button.
         // Note that we don't have the MB_SERVICE_NOTIFICATION flag.
         // This server is run attended, for simplicity.
         ::MessageBox(0, "Chat Server Currently Running...\n\n"
            "Press OK to end server.", szDiscussion, 
            MB_OK|MB_ICONINFORMATION);

         // Unregister the server from the Chat Broker 
         pBroker->Unregister(bstrDiscussion);
      } catch (_com_error & er) {
         if (er.Description().length()>0) {
            ::MessageBox(NULL, er.Description(), 
               TEXT("Error Description"), 
               MB_OK|MB_SERVICE_NOTIFICATION|MB_ICONERROR);
         } else {
            ::MessageBox(0, er.ErrorMessage(), 
               "ChatServer", 
               MB_OK|MB_SERVICE_NOTIFICATION|MB_ICONERROR);
         }
      }
      // OUR CODE----End


        // Generated code...Begin

        //  MSG msg;
        //  while (GetMessage(&msg, 0, 0, 0))
        //      DispatchMessage(&msg);
        _Module.RevokeClassObjects();
        Sleep(dwPause); //wait for any threads to finish

        // Generated code...End

      // OUR CODE----Begin
      // Unregister interface marshalers
      UnregisterInterfaceMarshaler(dwBroker);
      UnregisterInterfaceMarshaler(dwServer);
      UnregisterInterfaceMarshaler(dwClient);
      // OUR CODE----End
    }

    _Module.Term();
    CoUninitialize();
    return nRet;
}
